%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WN_CONNECT source v0.00 : wn_ic_measures module
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(wn_ic_measures, [
        wn_res/3,
        wn_jcn/3,
        wn_lin/3,
		lcs/4,
        information_content/3,
        frequency_of_use/2,
        gen_all_hyponyms_of/2,
        synset_tag_num/2
   ]).

:- use_module(wn_hypernyms).
:- use_module(wn_synsets).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% RESNIK SIMILARIRY MEASURE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_res(+Word1, +Word2, -Degree)
%%% This predicate implements the RESNIK similarity measure, based on information content.
%%% Takes two concepts (terms -- Word:SS_type:Sense_num) and returns the degree of
%%% similarity between them. Note that we do not explicitly require information about
%%% the synset type and sense number of a word. 
%%%
%%% We check that both Word1 and Word2 are nouns or verbs but not combinations of them.
%%%
%%% sim_RES(c1,c2)= IC(lcs(c1,c2))
%%%
%%% NOTE: "Word1:SS_type1:W1_Sense_num" denotes de concept c1 and "Word2:SS_type2:W2_Sense_num"
%%%       de concept c2.
%%%

wn_res(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree) :-
((nonvar(Word1), nonvar(Word2)) ->
  (member((SS_type1,SS_type2), [(n,n), (v,v)]) ->
    (var(W1_Sense_num) ->
        max_wordnet_sense(Word1, SS_type1, W1MaxSense),
        %% Nondeterministically generates all senses of Word1
        between(1, W1MaxSense, W1_Sense_num),
        (var(W2_Sense_num) ->
            max_wordnet_sense(Word2, SS_type2, W2MaxSense),
            %% Nondeterministically generates all senses of Word2
            between(1, W2MaxSense, W2_Sense_num),
            res(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
            ;
            res(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
        )
        ;
        res(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
    )
    ;
    write("ERROR: Both Word1 and Word2 must be either nouns or verbs (but not combinations of them or adjectives)"),
    nl
  )
  ;
  write("ERROR: Word1 or Word2 is a variable. You must enter a specific word (either noun or verb)."),
  nl
).

%%%%%%%%%%%%%%%%%%%
%%% res(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% Nondeterministic predicate. It obtains the degree of similarity of two concepts, associated
%%% to Word1 and Word2
%%%
%%% Degree = sim_RES(c1,c2)= IC(lcs(c1,c2))
%%%
%%%
res(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    lcs(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, LCS, Root_ID),
    information_content(LCS, Root_ID, Degree).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% JIANG & CONRATH SIMILARIRY MEASURE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_jcn(+Word1, +Word2, -Degree)
%%% This predicate implements the JIANG & CONRATH similarity measure, based on information
%%% content.
%%% Takes two concepts (terms -- Word:SS_type:Sense_num) and returns the degree of
%%% similarity between them. Note that we do not explicitly require information about
%%% the synset type and sense number of a word. 
%%%
%%% We check that both Word1 and Word2 are nouns or verbs but not combinations of them.
%%%
%%% sim_JCN(c1,c2)= 1/ [IC(c1) + IC(c2) - 2*IC(lcs(c1,c2))]
%%%
%%% NOTE: "Word1:SS_type1:W1_Sense_num" denotes de concept c1 and "Word2:SS_type2:W2_Sense_num"
%%%       de concept c2.
%%%

wn_jcn(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree) :-
((nonvar(Word1), nonvar(Word2)) ->
  (member((SS_type1,SS_type2), [(n,n), (v,v)]) ->
    (var(W1_Sense_num) ->
        max_wordnet_sense(Word1, SS_type1, W1MaxSense),
        %% Nondeterministically generates all senses of Word1
        between(1, W1MaxSense, W1_Sense_num),
        (var(W2_Sense_num) ->
            max_wordnet_sense(Word2, SS_type2, W2MaxSense),
            %% Nondeterministically generates all senses of Word2
            between(1, W2MaxSense, W2_Sense_num),
            jcn(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
            ;
            jcn(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
        )
        ;
        jcn(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
    )
    ;
    write("ERROR: Both Word1 and Word2 must be either nouns or verbs (but not combinations of them or adjectives)"),
    nl
  )
  ;
  write("ERROR: Word1 or Word2 is a variable. You must enter a specific word (either noun or verb)."),
  nl
).

%%%%%%%%%%%%%%%%%%%
%%% jcn(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% Nondeterministic predicate. It obtains the degree of similarity of two concepts, associated
%%% to Word1 and Word2
%%%
%%% Degree = sim_JCN(c1,c2)= 1/ [IC(c1) + IC(c2) - 2*IC(lcs(c1,c2))]
%%%
%%%
jcn(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    lcs(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, LCS, Root_ID),
    wn_synset_ID_of(Word1:SS_type:W1_Sense_num, W1_synset_ID),
    wn_synset_ID_of(Word2:SS_type:W2_Sense_num, W2_synset_ID),
    information_content(W1_synset_ID, Root_ID, W1_IC),
    information_content(W2_synset_ID, Root_ID, W2_IC),
    information_content(LCS, Root_ID, LCS_IC),
    ((W1_IC + W2_IC =:= 2*LCS_IC) ->
        %%% Zero distance between concepts W1 and W2 (usually W1 and W2 are the same word)
        %%% We use the concept of "almost zero distance" proposed by S. Patwardhan.
        frequency_of_use(Root_ID, F_root),
        Degree is 1/ (-log((F_root - 1)/ F_root))
        ;
        Degree is 1/ (W1_IC + W2_IC - 2*LCS_IC)
    ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% LIN SIMILARIRY MEASURE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_lin(+Word1, +Word2, -Degree)
%%% This predicate implements the JIANG & CONRATH similarity measure, based on information
%%% content.
%%% Takes two concepts (terms -- Word:SS_type:Sense_num) and returns the degree of
%%% similarity between them. Note that we do not explicitly require information about
%%% the synset type and sense number of a word. 
%%%
%%% We check that both Word1 and Word2 are nouns or verbs but not combinations of them.
%%%
%%% sim_JCN(c1,c2)= 1/ [IC(c1) + IC(c2) - 2*IC(lcs(c1,c2))]
%%%
%%% NOTE: "Word1:SS_type1:W1_Sense_num" denotes de concept c1 and "Word2:SS_type2:W2_Sense_num"
%%%       de concept c2.
%%%

wn_lin(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree) :-
((nonvar(Word1), nonvar(Word2)) ->
  (member((SS_type1,SS_type2), [(n,n), (v,v)]) ->
    (var(W1_Sense_num) ->
        max_wordnet_sense(Word1, SS_type1, W1MaxSense),
        %% Nondeterministically generates all senses of Word1
        between(1, W1MaxSense, W1_Sense_num),
        (var(W2_Sense_num) ->
            max_wordnet_sense(Word2, SS_type2, W2MaxSense),
            %% Nondeterministically generates all senses of Word2
            between(1, W2MaxSense, W2_Sense_num),
            lin(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
            ;
            lin(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
        )
        ;
        lin(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
    )
    ;
    write("ERROR: Both Word1 and Word2 must be either nouns or verbs (but not combinations of them or adjectives)"),
    nl
  )
  ;
  write("ERROR: Word1 or Word2 is a variable. You must enter a specific word (either noun or verb)."),
  nl
).

%%%%%%%%%%%%%%%%%%%
%%% lin(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% Nondeterministic predicate. It obtains the degree of similarity of two concepts, associated
%%% to Word1 and Word2
%%%
%%% Degree = sim_LIN(c1,c2)= [2*IC(lcs(c1,c2))] / [IC(c1) + IC(c2)]
%%%
%%%
lin(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    lcs(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, LCS, Root_ID),
    wn_synset_ID_of(Word1:SS_type:W1_Sense_num, W1_synset_ID),
    wn_synset_ID_of(Word2:SS_type:W2_Sense_num, W2_synset_ID),
    information_content(W1_synset_ID, Root_ID, W1_IC),
    information_content(W2_synset_ID, Root_ID, W2_IC),
    information_content(LCS, Root_ID, LCS_IC),
    ((W1_IC + W2_IC =:= 0) ->
        %%% If W1 and W2 don't have information content, we fix the degree of similarity to zero
        Degree = 0
        ;
        (LCS_IC =:= 0 ->
            %%% If LCS_IC doesn't have information content, is because the LCS is the root of the
            %%% hierarchy. This means that W1 and W2 are not related and we fix the degree of similarity
            %%% to zero
            Degree = 0
            ;
            Degree is (2*LCS_IC) / (W1_IC + W2_IC)
        )
    ).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATES FOR COMPUTING THE LESS COMMON SUBSUMER (LCS)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% lcs(+Word1, +Word2, -LCS, -Root)
%%% Calculates the Less Common Subsumer (LCS).
%%%
%%% It is an specialization of the predicate implemented in wn_sim_measures.pl
%%% The objective is to gain efficience in the obtention of the LCS.
%%%
%%% PROBLEMS WITH the Root of the hierarchy:
%%% Verbs do not have an explicit root. It may be that two verbs
%%% do not share a Less Common Subsumer. In this case LCS=Root and Root is assigned
%%% to a virtual root for verbs (Synset_ID = 100000000). Note that in this case, the
%%% information content of the LCS is 0 and the similarity of these two verbs should
%%% be around 0 also.
%%% On the other hand, nouns have a unique root hierarchy which is "entity" (synset_ID =
%%% 100001740). However, by uniformity of treatment we introduce a virtual root for names
%%% (Synset_ID = 100000000)
%%%
%%%
lcs(Word1, Word2, LCS, Root) :-
    wn_hypernyms(Word1, List_HyperNymSynSets1, Root),
    wn_hypernyms(Word2, List_HyperNymSynSets2, Root),
    find_lcs([Root|List_HyperNymSynSets1], [Root|List_HyperNymSynSets2], LCS).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% find_lcs(+List_HyperNymSynSets1, +List_HyperNymSynSets2, -LCS)
%%%
%%% List_HyperNymSynSets1 and List_HyperNymSynSets2 are lists of synset_IDs which are the hypernyms
%%% of the concepts (associated to the words) W1 and W2.
%%% LCS is the less common subsumer (a specific synset_ID).
%%%

find_lcs(L_SynSets1, L_SynSets2, LCS) :-
    find_lcs(L_SynSets1, L_SynSets2, 0, LCS).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% find_lcs(L_SynSets1, L_SynSets2, LCS_Acc, LCS).
%%%
%%% Compares two lists of (HyperNym) SynSets, element by element, until a mismatch is
%%% found. Then, the LCS are returned.
%%%
%%% LCS_Acc accumulate the last synset inspected.
%%% Initially LCS_Acc=0 (a virtual root synset).
%%%

find_lcs([], _L_SynSets2, LCS_Acc, LCS_Acc):-
    !.

find_lcs(_L_SynSets1, [], LCS_Acc, LCS_Acc):-
    !.

find_lcs([SS1|L_SynSets1], [SS2|L_SynSets2], _, LCS) :-
    SS1=:=SS2, !,
    find_lcs(L_SynSets1, L_SynSets2, SS1, LCS).

find_lcs(_L_SynSets1, _L_SynSets2, LCS_Acc, LCS_Acc).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%        PARAMETERS AND AUXILIARY PREDICATES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
maxDeep(n, 20).     % (max depth for nouns in WordNet HyperTrees is 20)
maxDeep(v, 14).     % (max depth for verbs in WordNet HyperTrees is 14)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% max_wordnet_sense(+Word, +Type, -MaxSense)
%%% Returns the maximum number of senses, 'MaxSense', of a word, 'Word', of type, 'Type'.
%%% If the parameters 'Word' and 'Type' are not instantiated then it returns the maximum number of senses
%%% for a word in the WordNet data base.
%%%
max_wordnet_sense(Word, WType, MaxSense) :-
    findall(WSense, wordnet:wn_s(_Synset_id, _WNum, Word, WType, WSense, _Tag_count), WSenseList),
    max_list(WSenseList, MaxSense).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% information_content(+Synset_ID, +Root_ID, -IC),
%%
%% Computes the information content IC of the concept denoted by the Synset_ID. This
%% quantity is defined as:
%%
%%    IC = -ln(Frequency/Frequency_Root)
%%
%%    if frequency_of_use(Synset_ID, Frequency) and frequency_of_use(Root_ID, Frequency_Root)
%%
%% NOTES:
%% 1) Root_ID is the synset number of the concept in the root of the hierarchy.
%% 2) IC(c) is defined as the natural logarithm of the probability of encountering an instance of
%%    a concept c (meassured in terms of a relative frequency of use of the concept c in a corpus).
%% 3) Natural logarithm:  logarithm to the base of the mathematical constant e.
%%

information_content(Synset_ID, Root_ID, IC) :-
    ( Synset_ID = 100000000 ->
            %% This case appears when comparing two nouns and the LCS = 100000000
            %% that is, it is the virtual root sysnset for nouns, with zero information content
            IC = 0
            ;
      (Synset_ID > 100000000, Synset_ID < 200000000) ->
            %% Synset_ID corresponds to a noun
            frequency_of_use(Synset_ID, F),
            %%% frequency_of_use of "entity" was computed and stored as a parameter
            %%% In this case, "Frequency_Root" is a non-zero value
            frequency_of_use(Root_ID, Frequency_Root),
            (F = 0 ->
             %% smoothing Frequency
             Frequency = 0.0000000001
             ;
             Frequency = F
            ),
            IC is -log(Frequency/Frequency_Root)
            ;
      Synset_ID = 200000000 ->
            %% This case appears when comparing two verbs and the LCS = 200000000
            %% that is, it the virtual root sysnset for verbs, with zero information content
            IC = 0
            ;
      (Synset_ID > 200000000, Synset_ID < 300000000) ->
            %% Synset_ID corresponds to a verb
            frequency_of_use(Synset_ID, F),
            frequency_of_use(Root_ID, Frequency_Root),
            (F = 0 ->
                %% smoothing Frequency
                Frequency = 0.0000000001
                ;
                Frequency = F
            ),
            IC is -log(Frequency/Frequency_Root)
    ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% frequency_of_use(+Synset_ID, -Frequency),
%%
%% In WordNet a synset represents a concept (which is identified by a Synset_ID). This
%% predicate computes the frequency of use of a concept in a corpus (using the information
%% stored in the last parameter of the predicate wn_s/6). We compute the "synset tag sum"
%% (see next predicate) of the concepts subsumed by Synset_ID adding them to obtain the
%% Frequency of the concept Synset_ID.
%%
%% Note that the concepts subsumed by Synset_ID are their hyponyms

%%% Frequency of use for the root concept "entity" (synset_ID = 100001740) -with_repetitions
frequency_of_use(100000000, 125518) :- !.
frequency_of_use(100001740, 125518) :- !.
%%% Frequency of use for the root concept "entity" (synset_ID = 100001740) - without_repetitions
%% rootFrequency(100001740, 94941).

%%% Frequency of use for the virtual root concept for verbs (Root = 200000000)
frequency_of_use(200000000, 95651) :- !.

frequency_of_use(Synset_ID, Frequency) :-
        frequency_of_use(with_repetitions, Synset_ID, Frequency).

%%% frequency_of_use(+Repetitions, +Synset_ID, -Frequency)
%%%
frequency_of_use(with_repetitions, Synset_ID, Frequency) :-
        gen_all_hyponyms_of(Synset_ID, List_all_Hyponym_IDs),
        sum_synset_tag_num([Synset_ID|List_all_Hyponym_IDs], Frequency).

frequency_of_use(without_repetitions, Synset_ID, Frequency) :-
        gen_all_hyponyms_of(Synset_ID, List_all_Hyponym_IDs),
        setof(X,member(X,List_all_Hyponym_IDs),List_all_Hyponym_IDs_NR),
        sum_synset_tag_num([Synset_ID|List_all_Hyponym_IDs_NR], Frequency).



         
%%% BEGIN DIFFERENCE LIST VERSION OF gen_all_hyponyms_of %%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% gen_all_hyponyms_of(+Synset_ID, -List_all_Hyponym_IDs).
%%%
%%% "List_all_Hyponym_IDs" is the list of all hyponyms of the synset "Synset_ID".
%%% The list "List_all_Hyponym_IDs" is a bag of synset_IDs hyponyms of "Synset_ID".
%%%
%%% NOTES: Because WordNet hierarchies are not actually trees, some synset nodes are
%%%        present several times. This can lead to overestimate some word frequencies
%%%        of use.
%%%
gen_all_hyponyms_of(Synset_ID, List_all_Hyponym_IDs) :-
        gen_all_hyponyms_of([Synset_ID], List_all_Hyponym_IDs, []).

%%% gen_all_hyponyms_of(+List_Synset_IDs, +List_all_Hyponym_IDs_Acc, -List_all_Hyponym_IDs).
%%%
gen_all_hyponyms_of([]) --> [].

gen_all_hyponyms_of([Synset_ID|List_Synset_IDs]) -->
        gen_hyponyms_of(Synset_ID, List_Hyponym_IDs),
        gen_all_hyponyms_of(List_Hyponym_IDs),
        gen_all_hyponyms_of(List_Synset_IDs).
        

%%% gen_hyponyms_of(+Synset_ID, -List_Hyponym_IDs)
%%%
gen_hyponyms_of(Synset_ID, List_Hyponym_IDs) -->
        {findall(Hyponym_ID, wordnet:wn_hyp(Hyponym_ID, Synset_ID), List_Hyponym_IDs)},
        List_Hyponym_IDs.

%%% END DIFFERENCE LIST VERSION OF gen_all_hyponyms_of %%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% synset_tag_num(+Synset_ID, -Frequency),
%%
%% The operator wn_s/6 stores the synset information of the WordNet database. Each word has
%% an entry:
%%            s(synset_ID,w_num,‘word’,ss_type,sense_number, tag_number).
%% where tag_number indicates how common a word is in relation to a text. The number is equal
%% to the times the word was found in a test corpus. Therefore the higher the number, the more
%% common the word. This parameter can be employed to obtain the frequency of use of a word
%% and if we sum the tag_number of all words in a synsent, we can obtain the specific frequency of
%% use associated to the whole synset.
%%
synset_tag_num(Synset_ID, Frequency) :-
integer(Synset_ID),
Synset_ID > 100000000,
Synset_ID < 500000000,
findall(Tag_number, wordnet:wn_s(Synset_ID, _, _, _, _, Tag_number), Word_frequencies),
sum(Word_frequencies, Frequency).

%%% sum(L, S), S is the sum of the list of numbers L.
%%%
sum(L, S) :- sum(L, 0, S).

sum([], A, A).
sum([N|Nums], A, S) :- NewA is N + A, sum(Nums, NewA, S).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% sum_synset_tag_num(+List_Synset_ID, -Frequency),
%%
sum_synset_tag_num([],0).
sum_synset_tag_num([Synset_ID|List_Synset_ID], Frequency) :-
        synset_tag_num(Synset_ID, FS),
        sum_synset_tag_num(List_Synset_ID, FLS),
        Frequency is FS + FLS.



