%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WN_CONNECT source v0.00 : wn_sim_measures module
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(wn_sim_measures, [
		wn_lcs/4,
      	wn_path/3,
		wn_path_nondet/3,
        wn_wup/3,
        wn_wup_nondet/3,
        wn_lch/3,
        wn_lch_nondet/3
   ]).

:- use_module(wn_hypernyms).
:- use_module(wn_synsets).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PATH SIMILARIRY MEASURE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_path(+Word1:SS_type1:W1_Sense_num, +Word2:SS_type2:W2_Sense_num, -Degree)
%%% This predicate implements the PATH similarity measure.
%%% Takes two concepts (terms -- Word:SS_type:Sense_num) and returns the degree of
%%% similarity between them. Note that we do not explicitly require information about
%%% the synset type and sense number of a word (that can be variables).
%%%
%%% We check that both Word1 and Word2 are nouns or verbs but not combinations of them.
%%%
%%% sim_PATH(c1, c2) = 1/len(c1, c2)
%%%
%%% NOTE: len(W1, W2) = (DepthW1-LCS_depth) + (DepthW2-LCS_depth) +1
%%%

wn_path(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree) :-
((nonvar(Word1), nonvar(Word2)) ->
  (utilities:member((SS_type1,SS_type2), [(n,n), (v,v)]) ->
    (var(W1_Sense_num) ->
        max_wordnet_sense(Word1, SS_type1, W1MaxSense),
        %% Nondeterministically generates all senses of Word1
        between(1, W1MaxSense, W1_Sense_num),
        (var(W2_Sense_num) ->
            max_wordnet_sense(Word2, SS_type2, W2MaxSense),
            %% Nondeterministically generates all senses of Word2
            between(1, W2MaxSense, W2_Sense_num),
            max_path(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
            ;
            max_path(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
        )
        ;
        max_path(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
    )
    ;
    write("ERROR: Both Word1 and Word2 must be either nouns or verbs (but not combinations of them or adjectives)"),
    nl
  )
  ;
  write("ERROR:  Word1 or Word2 is a variable. You must enter a specific word (either noun or verb)."),
  nl
).


%%%%%%%%%%%%%%%%%%%
%%% path(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% Nondeterministic predicate. It inspects a pair of HyperTrees associated to 
%%% Word1 and Word2 and obtains the degree of similarity between Word1 and Word2 
%%% (according to that pair of HyperTrees)
%%%
path(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    lcs(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, _, LCS_depth, DepthW1, DepthW2),
    Degree is (1 / ((DepthW1-LCS_depth) + (DepthW2-LCS_depth) +1)).


%%%%%%%%%%%%%%%%%%%
%%% max_path(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% When this predicate is executed, it is intended that all variables in 
%%% Word1:SS_type:W1_Sense_num and in Word2:SS_type:W2_Sense_num are completely 
%%% instantiated. 
%%%
%%% It obtains the degree of similarity between the concepts Word1:SS_type:W1_Sense_num 
%%% and Word2:SS_type:W2_Sense_num using the WUP measure.
%%%
%%% A concept can have more than one HyperTree, therefore several pairs of HyperTrees
%%% are possibly considered and a list of similarity degrees 'Degs' is obtained using wup/3.
%%% Finally, the maximum degree in the list 'Degs' is selected as a result.
%%%
max_path(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
	findall(Deg ,path(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Deg), Degs),
	utilities:max_list(Degs, Degree).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_path_nondet(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%%
%%% Nondeterministic predicate. It is the user interface to the local predicate path/3.
%%% 
%%% Inspects a pair of HyperTrees associated to Word1 and Word2 and obtains the degree
%%% of similarity between Word1 and Word2 (according to that pair of HyperTrees)
%%%
wn_path_nondet(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
   ( (SS_type=n; SS_type=v) ->
    path(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree)
    ;
	write("ERROR: Both Word1 and Word2 must be nouns or verbs but not combinations of them"),
	nl
   ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% WUP SIMILARIRY MEASURE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_wup(+Word1, +Word2, -Degree)
%%% This predicate implements the WUP similarity measure.
%%% Takes two concepts (terms -- Word:SS_type:Sense_num) and returns the degree of
%%% similarity between them. Note that we do not explicitly require information about
%%% the synset type and sense number of a word (that can be variables).
%%%
%%% We check that both Word1 and Word2 are nouns or verbs but not combinations of them.
%%%
%%% sim_WUP(c1,c2)= 2*depth(lcs(c1,c2)) / (Depth(c1)+Depth(c2))
%%%
%%%

wn_wup(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree) :-
((nonvar(Word1), nonvar(Word2)) ->
  (utilities:member((SS_type1,SS_type2), [(n,n), (v,v)]) ->
    (var(W1_Sense_num) ->
        max_wordnet_sense(Word1, SS_type1, W1MaxSense),
        %% Nondeterministically generates all senses of Word1
        between(1, W1MaxSense, W1_Sense_num),
        (var(W2_Sense_num) ->
            max_wordnet_sense(Word2, SS_type2, W2MaxSense),
            %% Nondeterministically generates all senses of Word2
            between(1, W2MaxSense, W2_Sense_num),
            max_wup(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
            ;
            max_wup(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
        )
        ;
        max_wup(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
    )
    ;
    write("ERROR: Both Word1 and Word2 must be either nouns or verbs (but not combinations of them or adjectives)"),
    nl
  )
  ;
  write("ERROR: Word1 or Word2 is a variable. You must enter a specific word (either noun or verb)."),
  nl
).


%%%%%%%%%%%%%%%%%%%
%%% wup(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% Nondeterministic predicate. It inspects a pair of HyperTrees associated to 
%%% Word1 and Word2 and obtains the degree of similarity between Word1 and Word2 
%%% (according to that pair of HyperTrees)
%%%
wup(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    lcs(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, _, LCS_depth, DepthW1, DepthW2),
    Degree is (2 * LCS_depth / (DepthW1 + DepthW2)).


%%%%%%%%%%%%%%%%%%%
%%% max_wup(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% When this predicate is executed, it is intended that all variables in 
%%% Word1:SS_type:W1_Sense_num and in Word2:SS_type:W2_Sense_num are completely 
%%% instantiated. 
%%%
%%% It obtains the degree of similarity between the concepts Word1:SS_type:W1_Sense_num 
%%% and Word2:SS_type:W2_Sense_num using the WUP measure.
%%%
%%% A concept can have more than one HyperTree, therefore several pairs of HyperTrees
%%% are possibly considered and a list of similarity degrees 'Degs' is obtained using wup/3.
%%% Finally, the maximum degree in the list 'Degs' is selected as a result.
%%%
max_wup(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    findall(Deg ,wup(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Deg), Degs),
    utilities:max_list(Degs, Degree).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_wup_nondet(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%%
%%% Nondeterministic predicate. It is the user interface to the local predicate wup/3.
%%% 
%%% Inspects a pair of HyperTrees associated to Word1 and Word2 and obtains the degree
%%% of similarity between Word1 and Word2 (according to that pair of HyperTrees)
%%%
wn_wup_nondet(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
 	( (SS_type=n; SS_type=v) ->
        wup(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree)
        ;
        write("ERROR: Both Word1 and Word2 must be nouns or verbs but not combinations of them"),
        nl
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% LCH SIMILARIRY MEASURE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_lch(+Word1, +Word2, -Degree)
%%% This predicate implements the LCH similarity measure.
%%% Takes two concepts (terms -- Word:SS_type:Sense_num) and returns the degree of
%%% similarity between them. Note that we do not explicitly require information about
%%% the synset type and sense number of a word  (that can be variables).
%%%
%%% We check that both Word1 and Word2 are nouns or verbs but not combinations of them.
%%%
%%% sim_LCH (c1, c2) = −ln[ len(c1,c2) / (2 * max{depth(c)|c in WordNet})]
%%%
%%% NOTE 1: len(W1, W2) = (DepthW1-LCS_depth) + (DepthW2-LCS_depth) +1
%%% NOTE 2: max{depth(c)|c in WordNet} is the maximum depth of a concept in the WordNet
%%%         data base. In practice, is a fixed constant for each part of speech
%%%           MaxDepth(n) = 20    (Nouns)
%%%           MaxDepth(v) = 14    (Verbs)
%%%

wn_lch(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree) :-
((nonvar(Word1), nonvar(Word2)) ->
  (utilities:member((SS_type1,SS_type2), [(n,n), (v,v)]) ->
    (var(W1_Sense_num) ->
        max_wordnet_sense(Word1, SS_type1, W1MaxSense),
        %% Nondeterministically generates all senses of Word1
        between(1, W1MaxSense, W1_Sense_num),
        (var(W2_Sense_num) ->
            max_wordnet_sense(Word2, SS_type2, W2MaxSense),
            %% Nondeterministically generates all senses of Word2
            between(1, W2MaxSense, W2_Sense_num),
            max_lch(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
            ;
            max_lch(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
        )
        ;
        max_lch(Word1:SS_type1:W1_Sense_num, Word2:SS_type2:W2_Sense_num, Degree)
    )
    ;
    write("ERROR: Both Word1 and Word2 must be either nouns or verbs (but not combinations of them or adjectives)"),
    nl
  )
  ;
  write("ERROR: Word1 or Word2 is a variable. You must enter a specific word (either noun or verb)."),
  nl
).


%%%%%%%%%%%%%%%%%%%
%%% lch(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% Nondeterministic predicate. It inspects a pair of HyperTrees associated to 
%%% Word1 and Word2 and obtains the degree of similarity between Word1 and Word2 
%%% (according to that pair of HyperTrees)
%%%
lch(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
    lcs(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, _, LCS_depth, DepthW1, DepthW2),
    Len_W1_W2 is ((DepthW1-LCS_depth) + (DepthW2-LCS_depth) + 1),
    maxDepth(SS_type, MaxDepth),
    Degree is ((-1) * log(Len_W1_W2 / (2 * MaxDepth))).

%%%%%%%%%%%%%%%%%%%
%%% max_lch(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%% When this predicate is executed, it is intended that all variables in 
%%% Word1:SS_type:W1_Sense_num and in Word2:SS_type:W2_Sense_num are completely 
%%% instantiated. 
%%%
%%% It obtains the degree of similarity between the concepts Word1:SS_type:W1_Sense_num 
%%% and Word2:SS_type:W2_Sense_num using the LCH measure.
%%%
%%% A concept can have more than one HyperTree, therefore several pairs of HyperTrees
%%% are possibly considered and a list of similarity degrees 'Degs' is obtained using lch/3.
%%% Finally, the maximum degree in the list 'Degs' is selected as a result.
%%%
max_lch(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
	findall(Deg ,lch(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Deg), Degs),
	utilities:max_list(Degs, Degree).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_lch_nondet(+Word1:SS_type:W1_Sense_num, +Word2:SS_type:W2_Sense_num, -Degree)
%%%
%%% Nondeterministic predicate. It is the user interface to the local predicate lch/3.
%%% 
%%% Inspects a pair of HyperTrees associated to Word1 and Word2 and obtains the degree
%%% of similarity between Word1 and Word2 (according to that pair of HyperTrees)
%%%
wn_lch_nondet(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree) :-
	( (SS_type=n; SS_type=v) ->
		lch(Word1:SS_type:W1_Sense_num, Word2:SS_type:W2_Sense_num, Degree)
		;
		write("Both Word1 and Word2 must be nouns or verbs but not combinations of them"),
		nl
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PREDICATES FOR COMPUTING THE LESS COMMON SUBSUMER (LCS)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% wn_lcs(+Word1, +Word2, -LCS, -LCS_depth)
%%% Returns the Less Common Subsumer LCS of two words and its depth LCS_depth from the
%%% root of the HyperTree
%%%
%%% NOTE:
%%% This predicate is nondeterministic.
%%% It can be used without specifying the Type and Sense of a Word, but in this case
%%% the LCS for all combinations of types and senses of these two words are obtained.
%%%
%%% If you want to obtain the LCS for two precise concepts introduce Word1:W1_Type:W1_Sense
%%% and Word2:W2_Type:W2_Sense
%%%
wn_lcs(Word1, Word2, LCS, LCS_depth) :-
    wn_hypernyms(Word1, List_HyperNymSynSets1),
    write(List_HyperNymSynSets1), nl,
    wn_hypernyms(Word2, List_HyperNymSynSets2),
    write(List_HyperNymSynSets2), nl,
    find_lcs(List_HyperNymSynSets1, List_HyperNymSynSets2, LCS, LCS_d, _, _),
    LCS_depth is LCS_d + 1.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% lcs(+Word1, +Word2, -LCS, -LCS_depth, -DepthW1, -DepthW2)
%%% This predicate, additionally, returns the depth of Word1 and Word2 by efficiency
%%% reasons. We want to go through the hyperonym lists only once, so we calculate these
%%% quantities at the same time we calculate the LCS.
%%%
%%% NOTE:
%%% Because wn_hypernyms/4 works with concepts, this predicate requires to introduce
%%% as parameters:  Word1:W1_Type:W1_Sense and Word2:W2_Type:W2_Sense
%%% instead of only Word1 and Word2. Note also that W1_Type, W1_Sense, W2_Type and W2_Sense
%%% can be variables.
%%%
lcs(Word1, Word2, LCS, LCS_depth, DepthW1, DepthW2) :-
    wn_hypernyms(Word1, List_HyperNymSynSets1, Virtual_Root),
    wn_hypernyms(Word2, List_HyperNymSynSets2, Virtual_Root),
    find_lcs([Virtual_Root|List_HyperNymSynSets1], [Virtual_Root|List_HyperNymSynSets2], LCS, LCS_depth, DepthW1, DepthW2).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% find_lcs(+List_HyperNymSynSets1, +List_HyperNymSynSets2, -LCS, -LCS_depth, -DepthW1, -DepthW2)
%%%
%%% List_HyperNymSynSets1 and List_HyperNymSynSets2 are lists of synset_IDs which are the hypernyms
%%% of the the concepts (associated to the words) W1 and W2.
%%% LCS is the less common subsumer (a specific synset_ID).
%%% LCS_depth is the depth of LCS, that is its length from the root of the HyperTree to the LCS.
%%% DepthW1 and DepthW2 are the depth of the concepts (associated to the words) W1 and W2.
%%%
%%% INITIALIZATION:
%%% LCS is initialized to zero (a virtual synset_ID).
%%% LCS_depth is initialized to 0 in order to consider the effect of adding a virtual root synset.
%%%

find_lcs(L_SynSets1, L_SynSets2, LCS, LCS_depth, DepthW1, DepthW2) :-
    find_lcs(L_SynSets1, L_SynSets2, 0, 0, LCS, LCS_depth, DepthW1, DepthW2).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% find_lcs(L_SynSets1, L_SynSets2, LCS_Acc, LCS_depth_Acc, LCS, LCS_depth).
%%%
%%% Compares two lists of (HyperNym) SynSets, element by element, until a mismatch is
%%% found. Then, the LCS and its depth (which where accumulated) are returned.
%%%
%%% LCS_Acc and LCS_depth_Acc accumulate the last synset inspected and its depth respectively.
%%% Initially LCS_Acc=0 (a virtual root synset) and LCS_depth_Acc=1 (the depth of the virtual root 0).
%%%

find_lcs([], L_SynSets2, LCS_Acc, LCS_depth_Acc, LCS_Acc, LCS_depth_Acc, DepthW1, DepthW2):-
    !,
    DepthW1 = LCS_depth_Acc,
    length(L_SynSets2, Length_L_SynSets2),
    DepthW2 is LCS_depth_Acc + Length_L_SynSets2.

find_lcs(L_SynSets1, [], LCS_Acc, LCS_depth_Acc, LCS_Acc, LCS_depth_Acc, DepthW1, DepthW2):-
    !,
    length(L_SynSets1, Length_L_SynSets1),
    DepthW1 is LCS_depth_Acc + Length_L_SynSets1,
    DepthW2 = LCS_depth_Acc.

find_lcs([SS1|L_SynSets1], [SS2|L_SynSets2], _, LCS_depth_Acc, LCS, LCS_depth, DepthW1, DepthW2) :-
    SS1=:=SS2, !,
    New_LCS_depth_Acc is LCS_depth_Acc +1,
    find_lcs(L_SynSets1, L_SynSets2, SS1, New_LCS_depth_Acc, LCS, LCS_depth, DepthW1, DepthW2).

find_lcs(L_SynSets1, L_SynSets2, LCS_Acc, LCS_depth_Acc, LCS_Acc, LCS_depth_Acc, DepthW1, DepthW2):-
    length(L_SynSets1, Length_L_SynSets1),
    DepthW1 is LCS_depth_Acc + Length_L_SynSets1,
    length(L_SynSets2, Length_L_SynSets2),
    DepthW2 is LCS_depth_Acc + Length_L_SynSets2.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%        PARAMETERS AND AUXILIARY PREDICATES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
           maxDepth(n, 20).     % (max depth for nouns in WordNet HyperTrees is 20)
           maxDepth(v, 14).     % (max depth for verbs in WordNet HyperTrees is 20)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% max_wordnet_sense(+Word, +Type, -MaxSense)
%%% Returns the maximum number of senses, 'MaxSense', of a word, 'Word', of type, 'Type'.
%%% If the parameters 'Word' and 'Type' are not instantiated then it returns the maximum number of senses
%%% for a word in the WordNet data base.
%%%
max_wordnet_sense(Word, WType, MaxSense) :-
    findall(WSense, wordnet:wn_s(_Synset_id, _WNum, Word, WType, WSense, _Tag_count), WSenseList),
    utilities:max_list(WSenseList, MaxSense).



